# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Extra

from .model import (
    AbiItem,
    Block,
    InternalTransaction,
    Log,
    Transaction,
    VBaseModel,
)
