# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra, Field

from .model import SettingsRegion, VBaseModel


class SettingsModel(VBaseModel):

    region: Optional[SettingsRegion] = Field(
        None,
        description="The region from where all the webhooks will be posted for this project",
    )
